
SORT_GUESSES = (
    'SND.DiffusionBValue',
    'SND.EchoTime',
    'SND.InversionTime',
    'SND.RepetitionTime',
    'SND.FlipAngle',
    'SND.AcquisitionTimeStamp',
    'AcquisitionNumber',
    'InstanceNumber',
)
'''The meta data keywords used when trying to guess the sorting order.
Keys that come earlier in the list are given higher priority.'''
